
# THIS FILE IS GENERATED FROM SETUP.PY
short_version = '0.0.6'
version = '0.0.6'
full_version = '0.0.6'
git_revision = 'f9f544fbfbb6913ac159a8b57777b133ccec6f66'
release = True

if not release:
    version = full_version
